/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef FOLDER_H
#define FOLDER_H

#include <clist.h>
#include <chash.h>

#include "cntree.h"
#include "libetpan.h"

#include "manager.h"

#include "folder-types.h"
#include "mailbox-types.h"

typedef clist XePFoldersList;

/*! Type of folder. */
typedef enum 
  {
    XEP_FOLDER_TYPE_REAL,
    XEP_FOLDER_TYPE_VIRTUAL,
    XEP_FOLDER_TYPE_MAILBOX
  } XePFolderType;

/*! A folder.
 *  This structure holds the messages.
 *  Each folder is managed by a different thread.
 *  A folder can be of one of the following types:
 *     - R-Folder, or real folder. It contains messages;
 *     - V-Folder, or virtual folder.  It contains dynamic search results;
 *     - M-Folder, or mailbox folder.  Just materializes a mailbox.
 */

struct XePFolder_s
{
  /* First member must be id, second must be path.
   * This is so the XePGtkUtilsTree can use a generic structure
   * wich two members are the first and second column of the tree
   */
  char *id;
  char *path;          /**< Path in the mailbox hieararchy, relative
		        *   to its toplever directory. Only for MH,
		        *   NNTP, IMAP and MBOX hierarchies.      */
  XePFolderType type;
  long uid;            /**< Unique Identifier                     */
  XePManagerThread thread;

  XePMailbox *mailbox; /**< Mailbox to which it is related.       */
  clist *rule_refs;    /**< List of processing rules              */

  struct mail_storage *ep_storage;     /**< The storage           */
  struct mail_folder  *ep_folder;      /**< The message folder    */
  struct mail_envelopes_tree *ep_tree; /**< The message tree      */
  struct mail_envelopes_list *ep_list; /**< The message list      */

  clist *folder_refs;    /**< VFolder: list of folders to seach   */
  clist *condition_refs; /**< VFolder: list of conditions         */
  uint32_t total;
  uint32_t unseen;
  uint32_t recent;
};

clist *
a_Folder_list_new(void);

clist *
a_Folder_list_new_copy(clist *folders);

/*! Releases resources from a folder list.
 *  \param folderlist the list of folders to be deleted.
 */
void
a_Folder_list_free(clist *folders);

void
a_Folder_list_free_copy(clist *folders);

void
a_Folder_list_copy(clist *source, clist *target);

XePFolder *
a_Folder_new(XePFolderType type);

XePFolder *
a_Folder_new_copy(XePFolder *folder);

void
a_Folder_free(XePFolder *folder);

void
a_Folder_free_copy(XePFolder *folder);

void
a_Folder_copy(XePFolder *source, XePFolder *target);

/*! Find the folder associated to the given Unique Identifier (uid)
 *  \param uid unique identifier
 *  \return the XePFolder associated to it, or NULL if there
 *          is no folder with this uid
 */
XePFolder *
a_Folder_get_from_uid(long uid);

XePFolderType
a_Folder_get_type(XePFolder *folder);

/*! Update folder's ID.
 *  As setting the mailbox or the path of the folder invalidates its
 *  id string, this function must be called explicitly to recreate it.
 *  \param folder the folder which ID string has to be updated. The
 *         ID string is the concatenation of mailbox's ID and
 *         folder's path with a colon between them.
 */
void
a_Folder_update_id(XePFolder *folder);

/*! Return the folder's ID 
 *  \param folder the folder which ID is sought
 */
const char *
a_Folder_get_id(XePFolder *folder);

/*! Set folder's path, of in case of a VFolder, its name 
 * \param folder the folder which path/name is to be set
 * \param path the path/name of the folder/VFolder
 */
void
a_Folder_set_path(XePFolder *folder, const char *path);

/*! Set folder's path, of in case of a VFolder, its name 
 * \param folder the folder which path/name is to be set
 * \param path the path/name of the folder/VFolder, folder
 *        will get ownership on path.
 */

void
a_Folder_set_path_p(XePFolder *folder, char *path);

const char *
a_Folder_get_path(XePFolder *folder);

void
a_Folder_set_mailbox(XePFolder *folder, XePMailbox *mailbox);

XePMailbox *
a_Folder_get_mailbox(XePFolder *folder);

const char *
a_Folder_get_short_name(XePFolder *folder);

void
a_Folder_set_folder_refs(XePFolder *folder, clist *folder_refs);

const clist *
a_Folder_get_folder_refs(XePFolder *folder);

void
a_Folder_set_condition_refs(XePFolder *folder, clist *condition_refs);

const clist *
a_Folder_get_condition_refs(XePFolder *folder);


/*! Create a new XePFolderId */
XePFolderRef *
a_Folder_ref_new(void);

XePFolderRef *
a_Folder_ref_new_copy(XePFolderRef *folder_ref);

/*! Release resources used by XePFolderId */
void
a_Folder_ref_free(XePFolderRef *folder_ref);

/*! Release resources used by XePFolderRef content */
void
a_Folder_ref_free_content(XePFolderRef *folder_ref);

clist *
a_Folder_ref_list_new(void);

/*! Release resources used by a list of XePFolderRef */
void
a_Folder_ref_list_free(clist *folder_refs);

chash *
a_Folder_ref_hash_new(void);

void
a_Folder_ref_hash_free(void);

chash *
a_Folder_ref_hash_get(void);

int
a_Folder_register(XePFolder *folder);

int
a_Folder_unregister(XePFolder *folder);

int
a_Folder_register_list(clist *folders);

int
a_Folder_unregister_list(clist *folders);

XePFolder *
a_Folder_get_from_ref(const char *ref);

/*! Create the rules links used by a folder.
 *  A folder first holds the identifier (XePProcessingRuleId) 
 *  for the processing rules it uses. This function use these
 *  identifiers to search the rules, get the pointer to the
 *  XePProcessingRule and store it. The identifiers are the
 *  freeed and NULLified.
 */
void
a_Folder_update_rules_links(XePFoldersList *folders);


int
a_Folder_ref_link_to_folder(XePFolderRef *folder_ref);

/*! Relink all folder references to the correspondig folder
 *  This functions updates the XePFolderRef::folder member
 *  of all known XePFolderRefs.
 */
void
a_Folder_ref_relink_all(void);

/*! Get the state of the thread associated with the folder
 *  \param folder an existing folder 
 *  \return the thread state
 */
XePManagerThreadState
a_Folder_get_thread_state(XePFolder *folder);

int
a_Folder_connect(XePFolder *folder);

int
a_Folder_disconnect(XePFolder *folder);

int
a_Folder_noop(XePFolder *folder);

int
a_Folder_expunge(XePFolder *folder);

/*! Sync folder content with mailbox 
 *  Eventually connect to the mailbox and retrieve content.
 *  At the end of the operation, folder should have the same information
 *  as the mailbox.
 */
int
a_Folder_sync(XePFolder *folder);

/*! Checkpoint folder 
 */
int
a_Folder_check(XePFolder *folder);

/*! Append a message to a folder
 *  \param XePFolder *folder
 *  \param data pointer to the rfc2822/822 message
 *  \param size the size of the message
 *  \return MAIL_NO_ERROR on success and MAIL_ERROR_XXX on error
 */
int
a_Folder_append(XePFolder *folder, const char *data, size_t size);

int
a_Folder_mailbox_connect(XePFolder *folder);

int
a_Folder_mailbox_disconnect(XePFolder *folder);

int
a_Folder_mailbox_list(XePFolder *folder, struct mail_list **mb_list);

int
a_Folder_mailbox_lsub(XePFolder *folder, struct mail_list **mb_list);


/*! Append a file to a folder in an asynchronous way then call
 *  the user provided function when finished.
 *  \param folder the target XePFolder
 *  \param filename the file containing the RFC2822/822 message to append
 *  \param user_cb the user callback function called when the operation
 *         finishes
 *  \param user_data the user data to pass to the user callback function
 */
void
a_Folder_append_file_async(XePFolder *folder, const char *filename,
			   XePManagerMsgCb user_cb, void *user_data);

void
a_Folder_disconnect_async(XePFolder *folder,
			  XePManagerMsgCb user_cb, void *user_data);

/*! Get total count of message tree nodes, including dummy messages
 *  this is used for the tree view because it needs to know the 
 *  total number of rows displayed
 */
uint32_t
a_Folder_tree_get_total_count(XePFolder *folder);

#endif
