/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h> /* unlink */
#include <gtk/gtk.h>

#include <charconv.h> /* From libetpan! */

#include "intl.h"

#include "x_alloc.h"
#include "logging.h"

#include "manager.h"
#include "message.h"

#include "ui.h"
#include "gtk-ui.h"

#include "gtk-messageview.h"
#include "gtk-utils.h"

#include "utils.h"

enum
  {
    MIME_TYPE_COL,
    MIME_FILENAME_COL,
    MIME_DESCRIPTION_COL,
    MIME_SIZE_COL,
    MIME_MIME_COL,
    MIME_COL_NB
    
  };


/* Static functions declarations */

static void
GtkMessageview_standalone_delete_cb(GtkWidget *widget, 
				    GdkEvent *event,
				    gpointer data);

static GtkTreeStore *
GtkMessageview_mime_tree_store_new(void);

static GtkWidget *
GtkMessageview_mime_tree_view_new(GtkTreeStore *treestore, gboolean editable);

static gboolean
GtkMessageview_mime_drag_data_received_cb(GtkWidget *widget,
					  GdkDragContext *context,
					  gint x,
					  gint y,
					  GtkSelectionData *selection_data,
					  guint info,
					  guint time,
					  gpointer data);

static GtkWidget *
GtkMessageview_mime_tree_menu_new(XePGTKMessageView *mview, gboolean editable);

static gboolean
GtkMessageview_mime_button_release_cb(GtkWidget *widget,
				      GdkEventButton *event,
				      gpointer data);

static void
GtkMessageview_mime_save_as_cb(gpointer data,
			       guint action,
			       GtkWidget *widget);

static void
GtkMessageview_mime_save_all_cb(gpointer data,
				guint action,
				GtkWidget *widget);
static void
GtkMessageview_mime_add_cb(gpointer data,
			   guint action,
			   GtkWidget *widget);
static int
GtkMessageview_mime_add_file_cb(void *amview, const char *filename);

static int
GtkMessageview_mime_add_file(XePGTKMessageView *mview, 
			     const char *filename,
			     struct mailmime *mime);

static void
GtkMessageview_mime_remove_cb(gpointer data,
			      guint action,
			      GtkWidget *widget);

static void
GtkMessageview_mime_edit_cb(gpointer data,
			    guint action,
			    GtkWidget *widget);

static void
GtkMessageview_mime_add_node(GtkTreeStore *store, 
			     GtkTreeIter *iter, 
			     struct mailmime *mime);

static void
GtkMessageview_show_text_part(XePGTKMessageView *mview, 
			      char *data, size_t len,
			      char *source_charset);

static GtkWidget *
GtkMessageview_inlined_icon_new(XePGTKMessageView *mview,
				GtkTextIter *iter,
				XePUtilsInlinedData *i_data);

static void
GtkMessageview_inlined_icon_destroy_cb(GObject *vbox, gpointer data);

static void
GtkMessageview_inlined_icon_save_as_cb(GtkWidget *widget,
				       GdkEventButton *event,
				       gpointer data);

/* Function definitions */

XePGTKMessageView *
a_GtkMessageview_create(XetPan *xetpan, 
			GtkWidget *parent,
			gboolean editable)
{
  XePGTKMessageView *mview;
  GtkWidget *pane;
  GtkWidget *treeview;
  GtkWidget *scrollw;
  GtkWidget *text;
  GtkTextBuffer *buffer;
  GtkTreeStore *treestore;

  mview = (XePGTKMessageView *) a_UI_messageview_new();
  if (!mview)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      return NULL;
    }
  mview->type = XETPAN_UI_GTK;
  mview->toplevel = NULL;
  mview->xetpan = xetpan;
  mview->uimain = NULL;
  mview->editable = editable;

  /* Vertical pane */
  pane = gtk_vpaned_new();
  gtk_container_add(GTK_CONTAINER(parent), pane);
  mview->widget = pane;

  a_GtkUtils_paned_shrink_control(pane, &(mview->pane_position));

  /* Mime scrollwindow */
  scrollw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrollw),
				      GTK_SHADOW_IN);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollw), 
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_AUTOMATIC);

  if (editable)
    {
      gtk_paned_pack2(GTK_PANED(pane), scrollw, FALSE, TRUE);
    }
  else
    {
      gtk_paned_add1(GTK_PANED(pane), scrollw);
    }
  mview->mime_widget = scrollw;

  /* If it is not a compose widget, do not show attachement pane until
   * we know there are really some attachments
   */

  if (editable)
    gtk_widget_show(mview->mime_widget);

  /* Mime tree store */
  treestore = GtkMessageview_mime_tree_store_new();
  mview->mime_store = treestore;

  /* Mime tree view */
  treeview = GtkMessageview_mime_tree_view_new(treestore, editable);
  mview->mime_view = treeview;
  gtk_widget_show(treeview);
  gtk_container_add(GTK_CONTAINER(scrollw), treeview);

  /* Mime tree context menu */
  mview->mime_menu = GtkMessageview_mime_tree_menu_new(mview, editable);

  /* Signal */
  g_signal_connect(G_OBJECT(treeview), "button-release-event",
		   G_CALLBACK(GtkMessageview_mime_button_release_cb),
		   mview);

  /* Body scrollwindow */
  scrollw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrollw),
				      GTK_SHADOW_IN);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollw), 
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_AUTOMATIC);

  if (editable)
    {
      gtk_paned_pack1(GTK_PANED(pane), scrollw, TRUE, TRUE);
    }
  else
    {
      gtk_paned_add2(GTK_PANED(pane), scrollw);
    }
  gtk_widget_show(scrollw);

  /* Body text view */
  text = a_GtkUtils_text_view_new(NULL, 0, editable);
  
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text));
  

  a_GtkUtils_text_buffer_create_tags(buffer);

  gtk_container_add(GTK_CONTAINER(scrollw), text);

  mview->text = text;
  mview->buffer = buffer;
  mview->message = NULL;
  mview->folder = NULL;


  if (editable)
    {
      /* Drag and Drop 
       * Drag: remove the item 
       * Drop: add an item
       */
      GtkTargetEntry target_entry = {"text/uri-list", 0, 1};
      gtk_drag_dest_set(treeview, GTK_DEST_DEFAULT_ALL, &target_entry, 1,
			GDK_ACTION_COPY);
      
      g_signal_connect(G_OBJECT(treeview), "drag_data_received",
		       G_CALLBACK(GtkMessageview_mime_drag_data_received_cb),
		       mview);
    }

  gtk_widget_show(text);

  return mview;
}

XePGTKMessageView *
a_GtkMessageview_create_standalone(XetPan *xetpan, char *terminal, int editable)
{
  GtkWidget *toplevel;
  XePGTKMessageView *mview;
  toplevel = a_GtkUtils_toplevel_new(terminal,
				     _("Message"), 
				     NULL,
				     460, 460);

  mview = a_GtkMessageview_create(xetpan, toplevel, editable);
  mview->toplevel = toplevel;

  /* Signal */
  /* FIXME: If XePMessage contains a msg that was forged, for example 
   * when the message viewer was called from the command line to view
   * a message, we should take care of destroying XePMessage::msg too.
   * Currently, there is a memory leak because of ignoring this.
   */

  g_signal_connect(G_OBJECT(toplevel), "delete-event",
		   G_CALLBACK(GtkMessageview_standalone_delete_cb), mview);

  a_GtkMessageview_show(mview);
  gtk_widget_show(toplevel);

  return mview;
}

void
a_GtkMessageview_free(XePGTKMessageView *mview)
{
  if (!mview)
    return;

  gtk_widget_destroy(mview->widget);

  if (mview->inlined_list)
    {
      a_Utils_inlined_list_free(mview->inlined_list);
      mview->inlined_list = NULL;
    }

  /* Currently, the only case where we need to free the message 
   * ourselves is when the XePMessage was forged from 
   * a file by using the -m switch. In that case, 
   * message->owner is set to one, which let us know that we
   * can free its resources.
   */
  if (mview->message && mview->message->owner)
    {
      a_Message_free(mview->message);
    }
  x_free(mview);
}

static void
GtkMessageview_standalone_delete_cb(GtkWidget *widget, 
				    GdkEvent *event,
				    gpointer data)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  XetPan *xetpan = (XetPan *) mview->xetpan;

  if (a_UI_messageview_close((XePUIMessageView *) mview))
    {
      a_XetPan_quit(xetpan);
    }
}

void
a_GtkMessageview_link_to_uimain(XePGTKMessageView *mview, XePUIMain *uimain)
{
  mview->uimain = uimain;
  mview->toplevel = (GtkWidget *) uimain->toplevel;
}

/* MIME view widget */

/* First column: mime/type
 * Second column: file name
 * Third column: description
 */
static GtkTreeStore *
GtkMessageview_mime_tree_store_new(void)
{
  GtkTreeStore *treestore;

  treestore = gtk_tree_store_new(MIME_COL_NB,
				 G_TYPE_STRING,
				 G_TYPE_STRING,
				 G_TYPE_STRING,
				 G_TYPE_STRING,
				 G_TYPE_POINTER);
  return treestore;
}

static GtkWidget *
GtkMessageview_mime_tree_view_new(GtkTreeStore *treestore, gboolean editable)
{
  GtkWidget *treeview;
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;

  treeview = gtk_tree_view_new_with_model(GTK_TREE_MODEL(treestore));
    
  /* The treeview holds a reference and is responsible for freeing
   * the treestore
   */
  g_object_unref(G_OBJECT(treestore));

  /* MIME Type name column */
  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("MIME Type"),
						    renderer,
						    "text",
						    MIME_TYPE_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Filename"),
						    renderer,
						    "text",
						    MIME_FILENAME_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Description"),
						    renderer,
						    "text",
						    MIME_DESCRIPTION_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Size"),
						    renderer,
						    "text",
						    MIME_SIZE_COL,
						    NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);

  gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), editable);

  return treeview;
}

static gboolean
GtkMessageview_mime_drag_data_received_cb(GtkWidget *widget,
					  GdkDragContext *context,
					  gint x,
					  gint y,
					  GtkSelectionData *selection_data,
					  guint info,
					  guint time,
					  gpointer data)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  char *filename;
  guchar *text;
  gint len = 0;
  int r = 0;

  if (selection_data->data == NULL)
    return TRUE;

  text = selection_data->data;
  while (len < selection_data->length && r == 0)
    {
      if (strncasecmp(text + len, "file://", strlen("file://")) == 0)
	{
	  int i = 0;
	  len += strlen("file://");
	  while ((len + i < selection_data->length) && 
		 (text[len + i] != '\r' && text[len + i] != '\n'))
	    {
	      i++;
	    }
	  if (i)
	    {
	      filename = x_alloc(i + 1, char);
	      memcpy(filename, text + len, i);
	      filename[i] = 0x00;
	      r = GtkMessageview_mime_add_file(mview, filename, mview->message->mime);
	      free(filename);
	    }
	  len += i;
	  while (len < selection_data->length && (text[len] == '\r' || text[len] == '\n'))
	    {
	      len++;
	    }
	}
      else
	{
	  r = -1;
	}
    }
  a_GtkMessageview_mime_view_show(mview, mview->message);
  return TRUE;
}

static GtkWidget *
GtkMessageview_mime_tree_menu_new(XePGTKMessageView *mview, gboolean editable)
{
  GtkAccelGroup *accel_group;
  GtkItemFactory *item_factory;
  GtkItemFactoryEntry view_menu[] = {
    { _("/Save as..."), NULL, GtkMessageview_mime_save_as_cb, 0, NULL},
    { _("/Save all..."), NULL, GtkMessageview_mime_save_all_cb, 0, NULL}
  };

  GtkItemFactoryEntry edit_menu[] = {
    { _("/Add..."), NULL, GtkMessageview_mime_add_cb, 0, NULL},
    { _("/Remove"), NULL, GtkMessageview_mime_remove_cb, 0, NULL},
    { _("/Edit..."), NULL, GtkMessageview_mime_edit_cb, 0, NULL}
  };
  int view_count = sizeof(view_menu)/sizeof(view_menu[0]);
  int edit_count = sizeof(view_menu)/sizeof(view_menu[0]);
  int count;
  const char *view_path = "<MIME_View>";
  const char *edit_path = "<MIME_Edit>";
  const char *path;
  GtkItemFactoryEntry *menu;

  if (editable)
    {
      menu = edit_menu;
      count = edit_count;
      path = edit_path;
    }
  else
    {
      menu = view_menu;
      count = view_count;
      path = view_path;
    }

  accel_group = gtk_accel_group_new();
  item_factory = gtk_item_factory_new(GTK_TYPE_MENU, path, accel_group);
  gtk_item_factory_create_items(item_factory, 
				count,
				menu, mview);

  return gtk_item_factory_get_widget(item_factory, path);
}

struct mailmime *
GtkMessageview_mime_get_selected(GtkTreeView *treeview)
{
  GtkTreeSelection *selection;
  GtkTreeIter treeiter;
  GtkTreeModel *treemodel;
  struct mailmime *mime = NULL;

  selection = gtk_tree_view_get_selection(treeview);
  if (gtk_tree_selection_get_selected(selection, &treemodel, &treeiter))
    {
      gtk_tree_model_get(treemodel, &treeiter, MIME_MIME_COL, &mime, -1);
    }

  return mime;
}

static gboolean
GtkMessageview_mime_button_release_cb(GtkWidget *widget,
				      GdkEventButton *event,
				      gpointer data)
{
  XePGTKMessageView *mview =  (XePGTKMessageView *) data;
  XePMessage *message = mview->message;

  if (event->button == 1 || event->button == 3)
    {
      struct mailmime *mime;
      mime = GtkMessageview_mime_get_selected(GTK_TREE_VIEW(mview->mime_view));
      if (mime)
	{
	  if (mview->editable)
	    {
	      if (message->selected_mime == message->default_mime)
		{
		  char *text;
		  text = a_GtkMessageview_get_text(mview);
		  a_GtkMessageview_default_part_set_text(mview, text,
							 (text ? 
							  strlen(text) : 0));

		}
	    }
	  mview->message->selected_mime = mime;
	  if (mview->message->async)
	    {
	      a_UI_messageview_mime_fetch_show_selected_async
		((XePUIMessageView *)mview, mview->message);
	    }
	  else
	    {
	      a_GtkMessageview_mime_show_selected(mview, mview->message);
	    }
	}
      /* TODO: button 3 should not start fetching the message part 
       * or change the selection
       */
      if (event->button == 3)
	{
	  gtk_menu_popup(GTK_MENU(mview->mime_menu), NULL, NULL, 
			 NULL, NULL,
			 event->button, event->time);
	}
    }
  if (event->button == 2 || event->button == 3)
    return TRUE;
  return FALSE;
}

static void
GtkMessageview_mime_save_as_cb(gpointer data,
			       guint action,
			       GtkWidget *widget)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  GtkWidget *toplevel;
  const char *filename;
  struct mailmime_single_fields fields;
  struct mailmime *mime;

  A_LOGGING_TRACE;

  mime = GtkMessageview_mime_get_selected(GTK_TREE_VIEW(mview->mime_view));
  if (mime)
    {
      memset(&fields, 0, sizeof(struct mailmime_single_fields));
      if (mime->mime_fields)
	{
	  mailmime_single_fields_init(&fields,
				      mime->mime_fields,
				      mime->content_type);
	}
      filename = fields.disposition_filename;
      
      toplevel = gtk_widget_get_toplevel(mview->widget);
      a_GtkUtils_file_selection(_("Save MIME Attachment As..."),
				toplevel,
				filename,
				(XePGTKUtilsFileSelectionCb)a_Message_data_save_as,
				mview->message);
    }
}

static void
GtkMessageview_mime_save_all_cb(gpointer data,
				guint action,
				GtkWidget *widget)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  GtkWidget *toplevel;

  A_LOGGING_TRACE;

  toplevel = gtk_widget_get_toplevel(mview->widget);
  a_GtkUtils_dir_selection(_("Select Directory for MIME Attachments..."),
			   toplevel,
			   NULL,
			   NULL,
			   mview->message);

  A_LOGGING_SYSTEM_ERROR("Saving all MIME attachments not implemented yet");
}

static void
GtkMessageview_mime_add_cb(gpointer data,
			   guint action,
			   GtkWidget *widget)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  GtkWidget *toplevel;

  A_LOGGING_TRACE;

  toplevel = gtk_widget_get_toplevel(mview->widget);
  a_GtkUtils_file_selection(_("Select a file to attach..."),
			    toplevel,
			    NULL,
			    GtkMessageview_mime_add_file_cb,
			    mview);
}

static int
GtkMessageview_mime_add_file_cb(void *amview, const char *filename)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;
  XePMessage *message = mview->message;
  int r;

  A_LOGGING_TRACE;

  r = GtkMessageview_mime_add_file(mview, filename, message->selected_mime);
  a_GtkMessageview_mime_view_show(mview, message);

  return r;
}

static int
GtkMessageview_mime_add_file(XePGTKMessageView *mview, 
			     const char *filename,
			     struct mailmime *mime)
{
  struct mailmime *new_mime;
  int r = -1;

  A_LOGGING_TRACE;

  if (filename == NULL)
    {
      return 0;
    }

  new_mime = a_Message_mime_new_from_file(filename, NULL,
					  MAILMIME_MECHANISM_ERROR);
  if (new_mime)
    {
      if (mime == NULL || mime == mview->message->default_mime)
	{
	  mime = mview->message->mime;
	}

      a_Message_mime_add_submime(mime, new_mime);
      r = 0;
    }

  return r;
}

static void
GtkMessageview_mime_remove_cb(gpointer data,
			      guint action,
			      GtkWidget *widget)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  XePMessage *message = mview->message;

  A_LOGGING_TRACE;

  if (message->selected_mime && 
      message->selected_mime != message->mime &&
      message->selected_mime != message->default_mime &&
      message->selected_mime != message->mime->msg_mime)
    {
      a_Message_mime_remove_selected(message);
      a_GtkMessageview_mime_view_show(mview, message);
    }
  else
    {
      A_LOGGING_ERROR(_("The default MIME part cannot be removed"));
    }
}

static void
GtkMessageview_mime_edit_cb(gpointer data,
			    guint action,
			    GtkWidget *widget)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  XePMessage *message = mview->message;

  A_LOGGING_TRACE;
}

void
a_GtkMessageview_message_clear(XePGTKMessageView *mview)
{
  gtk_widget_hide(mview->mime_widget);
  gtk_tree_store_clear(mview->mime_store);
  gtk_text_buffer_set_text(mview->buffer, "", -1);
  mview->muid = 0;
  a_Message_free(mview->message);
  mview->message = NULL;
}


/* Message view callbacks called after a thread finishes its work 
 * It will test if it is the latest message we expect in which case
 * it will register it, otherwise free the XePMessage of the ignored
 * message
 * Actually, this is UI type independent...
 */
void
a_GtkMessageview_finalize(XePGTKMessageView *mview, XePMessage *message)
{

  /* Ignore and discard all messages that are not the one we expected
   */
  if (mview->muid != message->muid)
    {
      a_Message_free(message);
      return;
    }

  x_free(message->data);
  message->data = NULL;
  message->len = 0;
  mview->message = message;
}

void
a_GtkMessageview_mime_view_block(XePGTKMessageView *mview,
				 XePMessage *message)
{
  if (mview->muid != message->muid)
    return;
    
  gtk_widget_set_sensitive(mview->mime_widget, FALSE);
}

void
a_GtkMessageview_mime_view_unblock(XePGTKMessageView *mview, 
				   XePMessage *message)
{
  gtk_widget_set_sensitive(mview->mime_widget, TRUE);
}

void
a_GtkMessageview_mime_view_show(XePGTKMessageView *mview, 
				XePMessage *message)
{
  mailmessage *msg = message->msg;
  struct mailmime *mime = message->mime;
  GtkTreeStore *mime_store = mview->mime_store;
  GtkTreeIter iter;

  /* Ignore if this is not the currently displayed message */
  if (mview->muid != message->muid)
    {
      A_LOGGING_DEBUG("Attempt to show the tree of a message not currently displayed");
      return;
    }

  gtk_tree_store_clear(mime_store);

  if (msg && mime && mime->msg_mime && mime->msg_mime->type != MAILMIME_SINGLE)
    {

      gtk_widget_show(mview->mime_widget);
     
      gtk_tree_store_append(mime_store, &iter, NULL);
      GtkMessageview_mime_add_node(mime_store, &iter, mime->msg_mime);
      gtk_tree_view_expand_all(GTK_TREE_VIEW(mview->mime_view));
      gtk_tree_view_columns_autosize(GTK_TREE_VIEW(mview->mime_view));
    }
  else if (mview->editable == 0)
    {
      gtk_widget_hide(mview->mime_widget);
    }
}

void
a_GtkMessageview_mime_show_selected(XePGTKMessageView *mview,
				    XePMessage *message)
{
  GtkTextBuffer *buffer = mview->buffer;
  GtkTextIter iter;
  
  gtk_text_buffer_set_text(buffer, "", -1);
  
  if (message->msg && mview->editable == 0)
    a_GtkMessageview_show_headers(mview, message);
  
  gtk_text_buffer_get_end_iter(buffer, &iter);
  
  if (message->async == 0)
    {
      a_Message_decode_selected(message);
    }

  if (message->data)
    {
      if (message->selected_mime &&
	  a_Message_mime_is_text(message->selected_mime))
	
	{
	  GtkMessageview_show_text_part(mview,
					message->data, message->len, 
					message->source_charset);
	}
      else
	{
	  char *mime_type;
	  mime_type = a_Utils_magic_data_mime_type(message->data, message->len);
	  
	  if (strncasecmp(mime_type, "text", 4) == 0)
	    {
	      char *source_charset;
	      const char *charset_tag = "charset=";

	      source_charset = strstr(mime_type, charset_tag);
	      if (source_charset)
		{
		  source_charset += strlen(charset_tag);
		}
	      GtkMessageview_show_text_part(mview, 
					    message->data, message->len,
					    source_charset);
	    }
	  else
	    {
	      if (strncasecmp(mime_type, "image", 5) == 0)
		{
		  a_GtkUtils_text_buffer_append_image_data(buffer, 
							   message->data, 
							   message->len);
		  
		}
	      else
		{
		  gtk_text_buffer_insert(buffer, 
					 &iter,
					 _("No visible text part found"),
					 -1);
		}
	    }
	  x_free(mime_type);
	}
    }
  else
    {
      gtk_text_buffer_insert(buffer, 
			     &iter,
			     _("<Empty>"),
			     -1);
    }
}

/* Message view functions */
static void
GtkMessageview_mime_add_node(GtkTreeStore *mime_store, 
			     GtkTreeIter *iter, 
			     struct mailmime *mime)
{
  char *mime_type;
  struct mailmime *a_mime = NULL;
  clistiter *mime_iter;
  struct mailmime_single_fields fields;
  const char *description;
  const char *filename;
  char size_buf[XEP_UTILS_SIZE_BUF_MAX];

  if (mime == NULL)
    return;

  memset(&fields, 0, sizeof(struct mailmime_single_fields));
  if (mime->mime_fields)
    {
      mailmime_single_fields_init(&fields,
				  mime->mime_fields,
				  mime->content_type);
    }

  description = fields.description;
  filename = fields.disposition_filename;
  a_Utils_format_size(size_buf, mime->length);
  mime_type = a_Message_mime_get_type(mime->content_type);

  gtk_tree_store_set(mime_store, iter, 
		     MIME_TYPE_COL, mime_type,
		     MIME_FILENAME_COL, (filename ? filename : ""),
		     MIME_DESCRIPTION_COL, (description ? description : ""),
		     MIME_SIZE_COL, size_buf,
		     MIME_MIME_COL, mime,
		     -1);
  x_free(mime_type);

  switch (mime->type)
    {
    case MAILMIME_SINGLE:
      break;
    case MAILMIME_MULTIPLE:
      for (mime_iter = clist_begin(mime->list); 
	   mime_iter; 
	   mime_iter = mime_iter->next)
	{
	  GtkTreeIter childiter;
	  gtk_tree_store_append(mime_store, &childiter, iter);
	  
	  a_mime = (struct mailmime *) mime_iter->data;
	  A_LOGGING_DEBUG("Adding sub node...");
	  GtkMessageview_mime_add_node(mime_store, &childiter, a_mime);
	  A_LOGGING_DEBUG("Sub node added");
	}
      break;
    case MAILMIME_MESSAGE:
      {
	GtkTreeIter childiter;
	gtk_tree_store_append(mime_store, &childiter, iter);
	A_LOGGING_DEBUG("-----");
	GtkMessageview_mime_add_node(mime_store, &childiter, mime->msg_mime);
	A_LOGGING_DEBUG("+++++");
      }
    }
}

void
a_GtkMessageview_show_headers(XePGTKMessageView *mview, XePMessage *message)
{
  GtkTextBuffer *buffer;
  GtkTextIter iter;

  mailmessage *msg;
  struct mailimf_single_fields *single_fields;

  const char *header_name;
  const char *header_name_tag;
  const char *header_value_tag;
  char *header_value;
  int headers[] = {
    XEP_MAIL_HEADER_SUBJECT,
    XEP_MAIL_HEADER_FROM,
    XEP_MAIL_HEADER_ORIG_DATE,
    XEP_MAIL_HEADER_TO,
    XEP_MAIL_HEADER_CC,
    XEP_MAIL_HEADER_REPLY_TO,
    XEP_MAIL_HEADER_OTHER
  };
  char *templates[] = {
    "%s",
    "%f <%F>",
    "%c", /* This is a strftime orig date template */
    "%t <%T>",
    "%c <%C>", /* This is a XetPan's CC template */
    "%r <%R>",
    NULL, /* Other header does not use a template */
    NULL};
  int i;


  msg           = message->msg;
  single_fields = &msg->single_fields;

  if (msg == NULL || single_fields == NULL)
    return;

  buffer = mview->buffer;

  gtk_text_buffer_get_end_iter(buffer, &iter);

  for (i = 0; i < sizeof(headers)/sizeof(headers[0]); i++)
    {
      header_value = a_Message_message_formatted_get_header(msg, 
							    headers[i],
							    "X-Mailer",
							    templates[i],
							    GTK_DEST_CHARSET);
      A_LOGGING_DEBUG("Header value = %s", header_value);
      if (header_value)
	{
	  header_name = a_Message_header_get_name_from_type(headers[i]);
	  a_GtkUtils_text_buffer_get_header_tag(headers[i],
						&header_name_tag,
						&header_value_tag);
						
	  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
						   header_name, -1,
						   header_name_tag, NULL);
	  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
						   header_value, -1,
						   header_value_tag, NULL);
	  gtk_text_buffer_insert(buffer, &iter, "\n", -1);

	  x_free(header_value);
	}
    }
  gtk_text_buffer_insert(buffer, &iter, "\n", -1);
}


static void
GtkMessageview_show_text_part(XePGTKMessageView *mview, 
			      char *data, size_t len,
			      char *source_charset)
{
  GtkTextBuffer *buffer = mview->buffer;
  GtkTextIter iter;
  const char *data_p;
  size_t remaining_len;
  int file_displayed = 0;
  clist *inlined_list;

  gtk_text_buffer_get_end_iter(buffer, &iter);

  /* Scan data for successive uuencoded or xxencoded data */

  remaining_len = len;
  data_p = data;

  /* Preprocess the body */
  /* a_Utils_filter(data, len, source_charset); */
  
  if (mview->inlined_list)
    {
      a_Utils_inlined_list_free(mview->inlined_list);
    }
  
  inlined_list = a_Utils_parse_body_text(data, len);
  mview->inlined_list = inlined_list;

  if (inlined_list)
    {
      clistiter *inlined_iter;
      XePUtilsInlinedData *i_data;

      for (inlined_iter = clist_begin(inlined_list); 
	   inlined_iter; 
	   inlined_iter = inlined_iter->next)
	{
	  i_data = (XePUtilsInlinedData *) inlined_iter->data;
	  switch (i_data->any.type)
	    {
	    case XEP_INLINED_DATA_TEXT:
	    case XEP_INLINED_DATA_CLEARSIGNED_TEXT:
	    case XEP_INLINED_DATA_ASCII_ARMORED_TEXT:
	      a_GtkUtils_text_buffer_convert_insert(buffer, &iter,
						    i_data->any.data,
						    i_data->any.size,
						    source_charset);
	      break;
	    case XEP_INLINED_DATA_UUENCODED:
	    case XEP_INLINED_DATA_YENCODED:
	    case XEP_INLINED_DATA_YENCODED2:
	    case XEP_INLINED_DATA_BINHEX:
	      file_displayed = 0;
	      if (a_GtkUtils_filename_is_supported_image(i_data->file.filename))
		{
		  char *tempfile;
		  char *decoded_file;
		  
		  tempfile = a_Utils_save_temp(i_data->any.data, 
					       i_data->any.size);
		  if (tempfile)
		    {
		      decoded_file = a_Utils_uulib_decode_file(tempfile,
							       NULL,
							       NULL,
							       NULL);
		      if (decoded_file && 
			  !a_GtkUtils_text_buffer_insert_image_file(buffer,
								    &iter,
								    decoded_file))
			
			{
			  file_displayed = 1;
			  unlink(decoded_file);
			  x_free(decoded_file);
			}
		      unlink(tempfile);
		      x_free(tempfile);
		    }
		}
	      if (!file_displayed)
		{
		  GtkWidget *uu_icon;
		  uu_icon = GtkMessageview_inlined_icon_new(mview, &iter, 
							    i_data);
		}
	      break;
	    case XEP_INLINED_DATA_NONE:
	    default:
	      A_LOGGING_SYSTEM_ERROR("Unknown inline data type");
	    }
	}
    }
}

static GtkWidget *
GtkMessageview_inlined_icon_new(XePGTKMessageView *mview,
				GtkTextIter *iter,
				XePUtilsInlinedData *inlined_data)
{
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *icon;
  GtkWidget *text;
  GtkWidget *event;
  char *tip_text;
  char msg_size[6];
  GtkTextChildAnchor *anchor;
  GtkTextBuffer *buffer;
  static GtkTooltips *tooltips = NULL;
  GtkTextView *textview;

  inlined_data->any.extra = mview;

  textview = GTK_TEXT_VIEW(mview->text);
  buffer = mview->buffer;

  anchor = gtk_text_buffer_create_child_anchor(buffer, iter);

  if (tooltips == NULL)
    {
      tooltips = gtk_tooltips_new();
      gtk_tooltips_enable(tooltips);
    }

  a_Utils_format_size(msg_size, inlined_data->any.size);
  tip_text = g_strdup_printf(_("Size: %s\nEncoding: %s"), msg_size,
			     a_Utils_inlined_data_type_as_string(inlined_data->any.type));

  vbox = gtk_vbox_new(FALSE, 0);
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  event = gtk_event_box_new();
  gtk_tooltips_set_tip(tooltips, event, tip_text, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), event, TRUE, FALSE, 0);

  icon = gtk_image_new_from_stock(GTK_STOCK_SAVE, GTK_ICON_SIZE_BUTTON);
  gtk_container_add(GTK_CONTAINER(event), icon);

  event = gtk_event_box_new();
  gtk_tooltips_set_tip(tooltips, event, tip_text, NULL);
  gtk_box_pack_start(GTK_BOX(vbox), event, FALSE, FALSE, 0);

  text = gtk_label_new(inlined_data->file.filename);
  gtk_label_set_selectable(GTK_LABEL(text), TRUE);
  gtk_container_add(GTK_CONTAINER(event), text);

  g_free(tip_text);

  gtk_widget_show_all(vbox);

  gtk_text_view_add_child_at_anchor(textview, vbox, anchor);

  /* Attach uuencoded data to vbox object */
  g_object_set_data(G_OBJECT(vbox), "inlined_data", inlined_data);

#if 0
  /* Signals: destroying the icon frees the related uuencoded data */
  g_signal_connect(G_OBJECT(vbox), "destroy", 
		   G_CALLBACK(GtkMessageview_inlined_icon_destroy_cb),
		   inlined_data);
#endif
  g_signal_connect(G_OBJECT(vbox), "button-release-event",
		   G_CALLBACK(GtkMessageview_inlined_icon_save_as_cb),
		   inlined_data);
  return vbox;  
}

static void
GtkMessageview_inlined_icon_destroy_cb(GObject *vbox, gpointer data)
{
  XePUtilsInlinedData *inlined_data = (XePUtilsInlinedData *) data;
  
  a_Utils_inlined_data_free(inlined_data);
}

static void
GtkMessageview_inlined_icon_save_as_cb(GtkWidget *widget,
				       GdkEventButton *event,
				       gpointer data)
{
  XePUtilsInlinedData *inlined_data = (XePUtilsInlinedData *) data;
  GtkWidget *toplevel;
  toplevel = 
    gtk_widget_get_toplevel(((XePGTKMessageView *) inlined_data->any.extra)->widget);

  a_GtkUtils_file_selection(_("Save Encoded Data As..."),
			    toplevel, 
			    inlined_data->file.filename,
			    (XePGTKUtilsFileSelectionCb)a_Utils_inlined_file_decode_as,
			    inlined_data);
}

gchar *
a_GtkMessageview_get_reply_text(XePGTKMessageView *mview)
{
  GtkTextIter start_iter;
  GtkTextIter end_iter;
  GtkTextTagTable *tag_table;
  GtkTextTag *reply_tag;
  gchar *reply_text = NULL;
  /* TODO */
  /* First check if there are "reply" tags, in which case, concatenate
   * all text marked with it, separating them with three empty lines
   * If no text is marked with "reply", check if there is some selection
   * and use it for a reply.
   * If not, use the whole text. 
   */

  /* Check for "reply-selection" marked text */

  tag_table = gtk_text_buffer_get_tag_table(mview->buffer);
  reply_tag = gtk_text_tag_table_lookup(tag_table, "reply-selection");
  gtk_text_buffer_get_bounds(mview->buffer, &start_iter, &end_iter);

  while (gtk_text_iter_forward_to_tag_toggle(&start_iter, reply_tag))
    {
      gchar *tmp;
      gchar *text;
      end_iter = start_iter;
      gtk_text_iter_forward_to_tag_toggle(&end_iter, reply_tag);
      text = gtk_text_buffer_get_text(mview->buffer, 
				      &start_iter, &end_iter,
				      FALSE);
      if (reply_text == NULL)
	{
	  reply_text = text;
	}
      else
	{
	  tmp = g_strconcat(reply_text, "\n\n\n", text, NULL);
	  g_free(reply_text);
	  g_free(text);
	  reply_text = tmp;
	}
      start_iter = end_iter;
    }

  if (reply_text)
    {
      return reply_text;
    }

  /* If no "reply-selection" found, check for selected text */
  if (gtk_text_buffer_get_selection_bounds(mview->buffer,
					   &start_iter,
					   &end_iter))
    {
      reply_text = gtk_text_buffer_get_text(mview->buffer,
					    &start_iter,
					    &end_iter,
					    FALSE);
    }

  return reply_text;
}

gchar *
a_GtkMessageview_get_text(XePGTKMessageView *mview)
{
  GtkTextIter start_iter;
  GtkTextIter end_iter;
  gchar *text;

  gtk_text_buffer_get_bounds(mview->buffer, &start_iter, &end_iter);
  text = gtk_text_buffer_get_text(mview->buffer, 
				  &start_iter, &end_iter,
				  FALSE);

  return text;  
}

void
a_GtkMessageview_default_part_set_text(XePGTKMessageView *mview,
				       char *text, size_t size)
{
  a_Message_mime_set_body_text(mview->message->default_mime,
			       text, size);
  if (a_Utils_has_8bit(text, 0))
    {
      a_Message_mime_remove_parameter(mview->message->default_mime, "charset");
      a_Message_mime_add_parameter(mview->message->default_mime,
				   "charset", GTK_DEST_CHARSET);
    }
}

#endif
