/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef MESSAGE_H
#define MESSAGE_H

#include <stdio.h> /* FILE */

#include <libetpan.h>

#include "folder-types.h"
#include "message-types.h"

/*! A structure containing all the data needed to display/edit a 
 *  message. Each XePMessage will contain a unique identifier (across
 *  a XetPan session).
 */
struct XePMessage_s
{
  char *data;             /**< Current in memory data */
  size_t len;             /**< Its size */
  XePFolder *folder;      /**< The related folder */
  uint32_t muid;          /**< A unique identifier for the structure,
			   *   * incremented for each call to
			   *   XePMessage_new(). Must be set and read
			   *   only from the main thread. */
  int async;              /**< Wether the message content is to be fetched */
  mailmessage *msg;       /**< The message structure. */
  struct mailmime *mime;  /**< The root MIME of the message */
  struct mailimf_fields *fields;  /**< The fields of the root MIME */
  struct mailimf_single_fields *single_fields;  
  struct mailmime *selected_mime; /**< The selected MIME */
  struct mailmime *default_mime; /**< The default MIME part */
  uint32_t flags_set;     /**< Flags to set, used by
			   *   a_Message_flags_change() */
  uint32_t flags_unset;   /**< Flags to unset, used by
			   *   a_Message_flags_change() */
  char *source_charset;   /**< The source charset of the data */
  const char *dest_charset; /**< The destination charset */
  int owner;              /**< Set to 1 if we should free everything
			   *   we point to (i.e. including msg and mime
			   */
  char *filename;         /**< Name of a file containing this message */
  int fd;                 /**< File descriptor related to filename */
  char *msg_data;         /**< Pointer to the in-memory RFC2822 message */
  size_t msg_size;        /**< Size of the in-memory RFC2822 message */
};


/*! Create a new XePMessage structure. This must be called from the
 *  main thread */
XePMessage *
a_Message_new(void);

/*! Create a new XePMessage that will own an in-memory RFC2822/822 message
 *  \param data pointer to the in-memory message
 *  \param size the size of the message
 *  \return a new XePMessage owning the in-memory data
 */
XePMessage *
a_Message_new_from_data_noparse(char *data, size_t size);

/*! Create a new XePMessage that is contained in a file.
 *  The file is not opened.
 *  \param filename name of the RFC2822/822 file
 *  \return a new XePMessage referencing the file
 */
XePMessage *
a_Message_new_from_file_noparse(const char *filename);

XePMessage *
a_Message_new_from_data(char *data, size_t len);

XePMessage *
a_Message_new_for_fetching(XePFolder *folder, mailmessage *msg);

/*! Create a XePMessage for a message to be composed
 *  The message is a message/rfc822 MIME message and
 *  the user agent header is set.
 */
XePMessage *
a_Message_new_to_compose(void);

XePMessage *
a_Message_new_to_reply(XePMessage *message);

XePMessage *
a_Message_new_to_forward(XePMessage *message);

XePMessage *
a_Message_new_to_redirect(XePMessage *message);

void
a_Message_free(XePMessage *message);


void
a_Message_free_cb(void *amessage, void *anargout);

XePMessage *
a_Message_clone(XePMessage *message);

/*! Open the file related to the message
 *  This file is set when XePMessage is created using 
 *  a_Message_new_from_file_noparse().
 *  \param message the XePMessage. Its filename member must point to
 *         an existing file.
 *  \return 0 on success and sets message's fd member to the file descriptor
 *          of the opened file. -1 on failure (no filename, or open error).
 */
int
a_Message_file_open(XePMessage *message);

/*! Close the file related to the message
 *  \param message the XePMessage. Its fd member should contain the file
 *         descriptor of the opened file. This one is set by 
 *         a_Message_file_open(). If it was a valid descriptor, the 
 *         file is closed and the descriptor is set to an invalid one (-1).
 */
void
a_Message_file_close(XePMessage *message);

void
a_Message_set_folder(XePMessage *message, XePFolder *folder);

XePFolder *
a_Message_get_folder(XePMessage *message);

char *
a_Message_message_formatted_get_header(mailmessage *msg, 
				       XePMessageHeaderType type,
				       char *other_header,
				       char *template,
				       char *dest_charset);

const char *
a_Message_header_get_name_from_type(XePMessageHeaderType type);

XePMessageHeaderType 
a_Message_header_get_type_from_name(const char *header_name);

/*! Remove all headers of some type
 *  This function iterates through the set of headers of a message
 *  and removes any occurence of header matching the provided header name
 *  \param message the message from which some headers are to be removed
 *  \param header_name the name of the header to remove from the message
 */
void
a_Message_header_remove_by_name(XePMessage *message, const char *header_name);

/*! Remove all headers of some type
 *  This function iterates through the set of headers of a message
 *  and removes any occurence of header matching the provided header type/name
 *  \param message the message from which some headers are to be removed
 *  \param type the XePMessageHeaderType of the headers to be removed
 *  \param header_name the name of the header to remove from the message in
 *         the case where the type is XEP_MAIL_HEADER_OTHER
 */
void
a_Message_header_remove(XePMessage *message,
			XePMessageHeaderType type,
			const char *other_header);


/*! Add a header specified by its name to the message 
 *  This function amounts to get the XePMessageHeaderType of the header
 *  then calling its sister function a_Message_header_add().
 *  \param message the message to which the header is added
 *  \param header_name the header name
 *  \param header_value the value of the header
 *  \param header_charset the character set in which the header value is 
 *         encoded
 */
void
a_Message_header_add_by_name(XePMessage *message,
			     const char *header_name,
			     const char *header_value,
			     const char *header_charset);

/*! Add a header to the message
 *  This function will call the specific a_Message_header_set/add_* function
 *  to set/add a header along with its value to the message
 *  \param message the message to which the header is added
 *  \param type the XePMessageHeaderType of the header
 *  \param header_name the header name in the case where the type is 
 *         XEP_MAIL_HEADER_OTHER
 *  \param header_value the value of the header
 *  \param header_charset the character set in which the header value is 
 *         encoded
 */
void
a_Message_header_add(XePMessage *message,
		     XePMessageHeaderType type,
		     const char *header_name,
		     const char *header_value,
		     const char *header_charset);

int
a_Message_header_set_from(XePMessage *message, 
			  const char *from,
			  const char *charset);
int
a_Message_header_add_to(XePMessage *message,
			const char *to, 
			const char *charset);

int
a_Message_header_add_cc(XePMessage *message, 
			const char *cc,
			const char *charset);

int
a_Message_header_add_bcc(XePMessage *message, 
			 const char *bcc,
			 const char *charset);

int
a_Message_header_add_reply_to(XePMessage *message,
			      const char *reply_to, 
			      const char *charset);

int
a_Message_header_add_other(XePMessage *message,
			   const char *header_name,
			   const char *header_value,
			   const char *header_charset);
int
a_Message_header_set_subject(XePMessage *message, 
			     const char *header_value, 
			     const char *header_charset);

char *
a_Message_message_formatted_get_local_date(struct mailimf_single_fields *fields,
					   char *template,
					   char *dest_charset);

char *
a_Message_message_formatted_get_orig_date(struct mailimf_single_fields *fields,
					  char *template,
					  char *dest_charset);
char *
a_Message_message_formatted_get_from(struct mailimf_single_fields *fields,
				     char *template,
				     char *dest_charset);
char *
a_Message_message_formatted_get_sender(struct mailimf_single_fields *fields,
				       char *template,
				       char *dest_charset);
char *
a_Message_message_formatted_get_reply_to(struct mailimf_single_fields *fields,
					 char *template,
					 char *dest_charset);
char *
a_Message_message_formatted_get_to(struct mailimf_single_fields *fields,
				   char *template,
				   char *dest_charset);
char *
a_Message_message_formatted_get_cc(struct mailimf_single_fields *fields,
				   char *template,
				   char *dest_charset);
char *
a_Message_message_formatted_get_bcc(struct mailimf_single_fields *fields,
				    char *template,
				    char *dest_charset);

char *
a_Message_message_formatted_get_message_id(struct mailimf_single_fields *fields,
					   char *template,
					   char *dest_charset);

char *
a_Message_message_formatted_get_in_reply_to(struct mailimf_single_fields *fields,
					    char *template,
					    char *dest_charset);

char *
a_Message_message_formatted_get_references(struct mailimf_single_fields *fields,
					   char *template,
					   char *dest_charset);

char *
a_Message_message_formatted_get_subject(struct mailimf_single_fields *fields,
					char *template,
					char *dest_charset);

char *
a_Message_message_formatted_get_comments(struct mailimf_single_fields *fields,
					 char *template,
					 char *dest_charset);

char *
a_Message_message_formatted_get_keywords(struct mailimf_single_fields *fields,
					 char *template,
					 char *dest_charset);

char *
a_Message_message_formatted_get_other(struct mailimf_single_fields *fields,
				      char *template,
				      char *other_header,
				      char *dest_charset);

int
a_Message_message_get_subject(struct mailimf_single_fields *fields,
			      char *dest_charset,
			      char **subject);

int
a_Message_message_get_from(struct mailimf_single_fields *fields,
			   char *dest_charset,
			   char **display_name,
			   char **email_address);

int
a_Message_message_get_other(struct mailimf_fields *fields,
			    char *other_header, 
			    char *dest_charset,
			    char **other);

void
a_Message_decode_selected(XePMessage *message);

/* Message fetching */

int
a_Message_get_bodystructure(XePMessage *message);

int
a_Message_mime_find_default(XePMessage *message);

int
a_Message_fetch_part(XePMessage *message);

int
a_Message_fetch(XePMessage *message);

int
a_Message_flags_change(XePMessage *message);

int
a_Message_mime_is_text(struct mailmime *mime);

char *
a_Message_mime_get_type(struct mailmime_content *content);

struct mailmime *
a_Message_mime_new_from_file(const char *filename, 
			     const char *mime_type,
			     int mechanism);

/*! Create a new MIME part from in-memory data
 *  \param data pointer to in-memory data. The struct mailmime will own
 *         the data.
 *  \param size the size of the in-memory data
 *  \param mime_type the MIME type of the data. NULL to automagically 
 *         guess the type from its content.
 *  \param mechanism the mechanism used for encoding.
 *  \return the newly created MIME part on success, and NULL on failure
 */
struct mailmime *
a_Message_mime_new_from_data(char *data,
			     size_t size,
			     const char *mime_type,
			     int mechanism);

void
a_Message_mime_set_body_text(struct mailmime *mime, 
			     char *data, 
			     size_t size);

struct mailmime *
a_Message_mime_new_multipart(char *subtype, int mechanism);

struct mailmime *
a_Message_mime_new_message(char *subtype, int mechanism);

struct mailmime *
a_Message_mime_add_part_to_selected(XePMessage *message, 
				    char *data,
				    size_t size,
				    const char *mime_type,
				    int mechanism);
int
a_Message_mime_add_parameter(struct mailmime *mime, 
			     const char *param_name,
			     const char *param_value);

/*! Remove all instances of a parameter in a MIME part
 *  \param mime the MIME part from which the parameter is to be removed
 *  \param param_mime the parameter to remove. If NULL, all parameters are
 *         removed
 *  \return 0 on success, or -1 if mime is NULL.
 */
int
a_Message_mime_remove_parameter(struct mailmime *mime,
				const char *param_name);

/*! Add a child MIME part to a MIME part
 *  If the parent MIME part is not a multipart MIME type, then the
 *  new MIME part will be added to a parent that is not of single type.
 *  \param mime the parent MIME part (or the older sibling)
 *  \param submime the MIME part to add
 *  \return 0 on success and 1 on failure
 */
int
a_Message_mime_add_submime(struct mailmime *mime, struct mailmime *submime);

/*! Add a description to a given MIME part
 *  \param mime the MIME part to describe
 *  \param description the description
 */
void
a_Message_mime_set_description(struct mailmime *mime, char *description);

/*! Remove the selected MIME part
 *  \param message the XePMessage. Its selected_part member must
 *         point to the MIME part to be removed
 *  \return 0 on success, -1 on failure.
 */
int
a_Message_mime_remove_selected(XePMessage *message);

/*! Get the first address specification of a From: field
 *  \param message the concerned XePMessage 
 *  \return a pointer to the string containing the address 
 *          specification in the the From: field (e.g.: joe@foo.bar). 
 *          The string must not be freed.
 */
const char *
a_Message_get_first_from_addr_spec(XePMessage *message);

/*! Get a specific header field
 *  \param message the XePMessage in which the field is to
 *         be searched
 *  \param type the XePMessageHeaderType of the header field
 *  \param other_header if type is XEP_MAIL_HEADER_OTHER, this
 *         argument contains the header looked for
 *  \return a pointer to the corresponding struct mailimf_field
 *          on success or NULL on failure
 */
struct mailimf_field *
a_Message_get_field(XePMessage *message,
		    XePMessageHeaderType type,
		    const char *other_header);

int
a_Message_data_save_as(XePMessage *message, const char *filename);

void
a_Message_write_to_file(XePMessage *message, FILE *file);

char *
a_Message_write_to_temp_file(XePMessage *message);

/*! Append a message to a folder
 *  This function is called by the manager upon XEP_MANAGER_APPEND
 *  request.
 *  \param message the XePMessage to append. Its folder member points
 *         to the target XePFolder, its filename member references 
 *         the file containing the data, or its msg_data member must
 *         point to in-memory message.
 */
void
a_Message_append_to_folder(XePMessage *message);

#endif
