/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_TTYUI

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <ncurses.h>
#include <sys/poll.h>

#include "logging.h"
#include "defines.h"
#include "x_alloc.h"
#include "xetpan.h"
#include "ui.h"
#include "tty-ui.h"
#include "remote.h"

typedef struct XePTTYUI_s XePTTYUI;

/*! A structure containging global TTY data 
 *  When the default UI is TTY, this holds the data concerning
 *  the event loop, the watched file descriptors etc.
 */

struct XePTTYUI_s
{
  struct pollfd *fds;
  unsigned int nfds;
  unsigned int fds_size;
  int n_events; /* Used when a_TtyUI_events_pending() is called */
};

typedef struct XePTTYScr_s XePTTYScr;

struct XePTTYScr_s
{
  SCREEN *scr;
  char *tty;
  char *term;
  FILE *in;
  FILE*out;
};

typedef struct XePTTYPython_s XePTTYPython;

struct XePTTYPython_s
{
  char *term;
  char *tty;
};

static void
TtyUI_python_console(XePUIMain *uimain);

static XePTTYPython *
TtyPython_console_new(XetPan *xetpan, char *terminal);

static void
TtyPython_console_show(XePTTYPython *xeppy);


static short int
TtyUI_condition_set(short int xcondition);

static short int
TtyUI_condition_get(short int pcondition);

static void
TtyUI_io_info(XePUI *ui);

/***************************************************************************/
/*                                                                         */
/* Basic UI functions                                                      */
/*                                                                         */
/***************************************************************************/

/* Init display and abort if fail */
void
a_TtyUI_init(XetPan *xetpan, int *argc, char ***argv)
{
  A_LOGGING_TRACE;

  if (a_TtyUI_init_check(xetpan, argc, argv))
    {
      A_LOGGING_ERROR("TTY Terminal User Interface could not be initialized");
      exit(1);
    }
}

int
a_TtyUI_init_check(XetPan *xetpan, int *argc, char ***argv)
{
  XePTTYUI *ttyui;
  int ret = 0;

  A_LOGGING_TRACE;
  if (xetpan->ui->tty)
    {
      A_LOGGING_INFO("TTY Terminal User Interface  already initialized.");
    }
  else
    {
      if (isatty(0) == 0 || isatty(1) == 0)
	{
	  A_LOGGING_ERROR("This is not a terminal.");
	  ret = -1;
	}
      else 
	{
	  ttyui =  x_alloc(1, XePTTYUI);
	  if (ttyui == NULL)
	    {
	      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
	      ret = -1;
	    }
	  else
	    {
	      ttyui->fds = x_alloc(10, struct pollfd);
	      if (ttyui->fds == NULL)
		{
		  A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
		  x_free(ttyui);
		  ret = -1;
		}
	      else
		{
		  ttyui->fds_size = 10;
		  ttyui->nfds = 0;
		  xetpan->ui->tty = ttyui;
		  A_LOGGING_INFO("Initialized TTY Terminal User Interface");
		}
	    }
	}
    }
  return ret;
}


/*! Main TTY loop 
 *  Poll for the watched file descriptors.
 *  Then create the list of the callbacks to be called
 *  Call successively the callbacks using the privately copied list.
 *  No callback should change the pointer to XetPan *xetpan.
 */

void
a_TtyUI_main(XetPan *xetpan)
{
  XePTTYUI *ttyui;

  A_LOGGING_TRACE;
  
  ttyui = (XePTTYUI *) xetpan->ui->tty;

  while (ttyui && ttyui->nfds > 0)
    {
      A_LOGGING_DEBUG("polling for %d file descriptors.", ttyui->nfds);
      ttyui->n_events = poll(ttyui->fds, ttyui->nfds, -1);
      A_LOGGING_DEBUG("poll returned %d events", ttyui->n_events);

      a_TtyUI_main_iteration(xetpan);
      ttyui = (XePTTYUI *) xetpan->ui->tty;
    }
  
  A_LOGGING_DEBUG("ended");

  x_free(ttyui->fds);
  x_free(ttyui);
  xetpan->ui->tty = NULL;
  A_LOGGING_DEBUG("xetpan->ui->tty = %p", xetpan->ui->tty);

}

/*! Return the pollfd.event equivalent to the XETPAN_POLL... */
static short int
TtyUI_condition_set(short int condition)
{
  int pcondition = 0;
  if (condition & XETPAN_POLLIN)
    pcondition |= POLLIN;
  if (condition & XETPAN_POLLPRI)
    pcondition |= POLLPRI;
  if (condition & XETPAN_POLLOUT)
    pcondition |= POLLOUT;
  if (condition & XETPAN_POLLERR)
    pcondition |= POLLERR;
  if (condition & XETPAN_POLLHUP)
    pcondition |= POLLHUP;
  if (condition & XETPAN_POLLNVAL)
    pcondition |= POLLNVAL;

  return pcondition;
}

/*! Return the XETPAN_POLL* equivalent */
static short int
TtyUI_condition_get(short int pcondition)
{
  int condition = 0;
  if (pcondition & POLLIN)
    condition |= XETPAN_POLLIN;
  if (pcondition & POLLPRI)
    condition |= XETPAN_POLLPRI;
  if (pcondition & POLLOUT)
    condition |= XETPAN_POLLOUT;
  if (pcondition & POLLERR)
    condition |= XETPAN_POLLERR;
  if (pcondition & POLLHUP)
    condition |= XETPAN_POLLHUP;
  if (pcondition & POLLNVAL)
    condition |= XETPAN_POLLNVAL;

  return condition;
}

void
a_TtyUI_main_quit(XetPan *xetpan)
{
  XePTTYUI *ttyui = (XePTTYUI *) xetpan->ui->tty;

  A_LOGGING_TRACE;

  ttyui->nfds = 0;
}

int
a_TtyUI_events_pending(XetPan *xetpan)
{
  XePTTYUI *ttyui;
  A_LOGGING_TRACE;

  ttyui = (XePTTYUI *) xetpan->ui->tty;
  
  ttyui->n_events = poll(ttyui->fds, ttyui->nfds, 0);

  return (ttyui->n_events != 0);
}

void
a_TtyUI_main_iteration(XetPan *xetpan)
{
  XePTTYUI *ttyui;
  unsigned int fd_idx, active_idx;
  struct pollfd *fds;
  int i, n_events;
  int rcondition;
  carray *watched_fd;
  carray *active_watchers;
  XePUIIOWatcher *watcher, *watcher_copy;

  ttyui = (XePTTYUI *) xetpan->ui->tty;

  n_events = ttyui->n_events;

  if (n_events < 0)
    { /* Poll error */
      A_LOGGING_SYSTEM_ERROR("poll: %s", strerror(errno));
    }
  else if (n_events == 0)
    {
      /* Poll timed out. Just ignore for now. */
    }
  else
    { /* Poll succeeded. Look for callbacks */
      watched_fd = xetpan->ui->watched_fd;
      
      active_watchers = carray_new(n_events);
      A_LOGGING_DEBUG("ttyui->nfds = %d", ttyui->nfds);
      
      for (i = 0; i < n_events; i++)
	{
	  for (fd_idx = 0; fd_idx < ttyui->nfds; fd_idx++)
	    {
	      watcher = (XePUIIOWatcher *) carray_get(watched_fd, fd_idx);
	      fds = &(ttyui->fds[(int)(watcher->channel)]);
	      rcondition = TtyUI_condition_get(fds->revents);
	      if (rcondition & watcher->condition)
		{
		  A_LOGGING_DEBUG("found corresponding watcher: %d fd: %d",
				  watcher->watch_id,
				  watcher->fd);
		  watcher_copy = (XePUIIOWatcher *) 
		    malloc(sizeof(XePUIIOWatcher));
		  memcpy(watcher_copy, watcher, sizeof(XePUIIOWatcher));
		  watcher_copy->condition = rcondition;
		  carray_add(active_watchers, watcher_copy, &active_idx);
		}  
	    }
	}
      active_idx = carray_count(active_watchers);
      if (n_events != active_idx)
	{
	  A_LOGGING_SYSTEM_ERROR("Number of active fd (%d) differen from "
				 "the expected one (%d).",
				 n_events, active_idx);
	}
      else
	{
	  /* Now that we have a private copy of the needed data, we
	     can call safely the different callback. */
	  
	  for (i = 0; i < n_events; i++)
	    {
	      watcher_copy = (XePUIIOWatcher *) carray_get(active_watchers,
							   i);
	      /* FIXME: check if the call back want to remove itself */
	      watcher_copy->callback(watcher_copy->fd, 
				     watcher_copy->condition,
				     watcher_copy->data);
	      x_free(watcher_copy);
	    }
	  carray_free(active_watchers);
	}
    }
}

/* FIXME: This one is just for testing purposes */
static void
TtyUI_basic_keyboard_cb(int fd, short int condition, void *data)
{
  int ch;
  XePUIMain *uimain = (XePUIMain *) data;
  XetPan *xetpan = (XetPan *) uimain->xetpan;

  A_LOGGING_DEBUG("condition 0x%x", condition);

  set_term(((XePTTYScr *)uimain->toplevel)->scr);
  if (condition & (XETPAN_POLLHUP | XETPAN_POLLERR))
    {
      if (a_UI_mainview_close(uimain))
	a_XetPan_quit(xetpan);
    }
  else
    {
      ch = getch();
      switch (ch)
	{
	case KEY_REFRESH:
	case 0x0C:
	  A_LOGGING_DEBUG("Screen refresh");
	  refresh();
	  break;
	case KEY_RESIZE:
	  A_LOGGING_DEBUG("Screen resized");
	  refresh();
	  break;
	case 'q':
	  if (a_UI_mainview_close(uimain))
	    a_XetPan_quit(xetpan);
	  break;
	case 'P':
	  TtyUI_python_console(uimain);
	  break;
	default:
	  addch((chtype) ch);
	  refresh();
	}
    }
}

/******************************************************************************/
/*                                                                            */
/* UI screens                                                                 */
/*                                                                            */
/******************************************************************************/

void
a_TtyUI_first_launch(XetPan *xetpan)
{
  A_LOGGING_TRACE;
}

void
a_TtyUI_compose_pending_requests(XetPan *xetpan)
{
  A_LOGGING_TRACE;
}

XePUIMain *
a_TtyUI_mainview_create(XetPan *xetpan, char *terminal)
{
  XePUIMain *uimain;
  XePTTYFoldersView *fview;
  XePTTYSummaryView *sview;
  XePTTYMessageView *mview;
  char *tty  = NULL;
  char *term = NULL;
  XePTTYScr *toplevel;

  A_LOGGING_TRACE;

  uimain = a_UI_mainview_new();
  if (uimain == NULL)
    return NULL;

  toplevel = x_alloc(1, XePTTYScr);
  if (toplevel == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      x_free(uimain);
      return NULL;
    }

  uimain->type = XETPAN_UI_TTY;
  uimain->xetpan = xetpan;
  
  if (terminal == NULL)
    {
      tty = strdup(ttyname(0));
      if (getenv("TERM"))
	term = strdup(getenv("TERM"));
    }
  else
    {
      char *separator;
      int len;
      separator = strchr(terminal, '\t');
      if (separator == NULL)
	{
	  A_LOGGING_ERROR("Invalid terminal description: %s", terminal);
	  x_free(uimain);
	  return NULL;
	}
      len = separator - terminal;  /* Do not count terminal NULL */
      term = x_alloc(len + 1, char);
      term[0] = 0x00;
      strncat(term, terminal, len);
      len = strlen(terminal) - len - 1; /* Count separator */
      tty = x_alloc(len + 1, char);
      tty[0] = 0x00;
      strncat(tty, separator + 1, len);
    }

  toplevel->in = fopen(tty, "r");
  if (!toplevel->in)
    {
      A_LOGGING_SYSTEM_ERROR("fopen: %s", strerror(errno));
      x_free(uimain);
      x_free(toplevel);
      return NULL;
    }
  
  toplevel->out = fopen(tty, "w");
  if (!toplevel->out)
    {
      A_LOGGING_SYSTEM_ERROR("fopen: %s", strerror(errno));

      fclose(toplevel->in);
      x_free(uimain);
      x_free(toplevel);
      return NULL;
    }

  toplevel->term = term;
  toplevel->tty  = tty;
  toplevel->scr = newterm(term, toplevel->out, toplevel->in);

  if (toplevel->scr == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("newterm: %s", strerror(errno));

      fclose(toplevel->in);
      fclose(toplevel->out);
      x_free(toplevel->term);
      x_free(toplevel->tty);
      x_free(toplevel);
      x_free(uimain);
      return NULL;
    }

  uimain->toplevel = toplevel;

  /* FIXME: Will go in a separate functions */
  {
    set_term(toplevel->scr);  
    def_prog_mode();
    cbreak();
    keypad(stdscr, TRUE);
    curs_set(0);
    noecho();
  }
  {
    mvaddstr(4,0,"Subject");
    mvaddstr(4, COLS/2, "From");
    mvaddstr(4, COLS-10, "Date");
    mvchgat (4, 0, -1, A_REVERSE, 0, NULL);
    
    mvaddstr(LINES-2, 0, "View:");
    mvaddstr(LINES-2, COLS/2, "Folder:");
    mvaddstr(LINES-2, COLS-0, "0");
    mvchgat (LINES-2, 0, -1, A_REVERSE, 0, NULL);

    mvaddstr(LINES-1, 0, "XetPan:");
  }

#if 0
  {
    int fd_mon;
    int child_status;
    a_Utils_exec_cmd("vim", 
		     fileno(toplevel->in), 
		     fileno(toplevel->out), 
		     fileno(toplevel->out), &fd_mon, NULL,
		     0);

    read(fd_mon, (char *)&child_status, sizeof(int));
    printf("Status = %d\n", child_status);
    getc(toplevel->in);
  }
#endif
  refresh();
  A_LOGGING_DEBUG("TTY UI term = %s tty = %s in = %d", toplevel->term, 
		  toplevel->tty, fileno(toplevel->in));

  uimain->kb_watch_id = a_UI_add_io_watch(xetpan, fileno(toplevel->in), 
					  XETPAN_POLLIN | XETPAN_POLLHUP,
					  TtyUI_basic_keyboard_cb, uimain);
  return uimain;
}

void
a_TtyUI_mainview_close(XePUIMain *uimain)
{
  XePTTYScr *toplevel;

  A_LOGGING_TRACE;

  if (uimain->type != XETPAN_UI_TTY)
    return;

  a_UI_remove_io_watch((XetPan *) uimain->xetpan, uimain->kb_watch_id);

  if (uimain->remote)
    {
      a_Remote_send_cmd_result((XePRemote *) uimain->remote);
    }

  toplevel = (XePTTYScr *) uimain->toplevel;

  set_term(toplevel->scr);
  endwin();
  delscreen(toplevel->scr);
  
  x_free(toplevel->tty);

  x_free(toplevel->term);
  
  fclose(toplevel->in);
  fclose(toplevel->out);
  x_free(toplevel);
  x_free(uimain);
}

void
a_TtyUI_folder_updated(XePUIMain *uimain, XePFolder *folder)
{
  A_LOGGING_DEBUG("folder '%s'\n", a_Folder_get_id(folder));
  A_LOGGING_SYSTEM_ERROR("not implemented yet");
}

void
a_TtyUI_messageview_finalize(XePUIMessageView *mview, XePMessage *message)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");
}

void
a_TtyUI_messageview_mime_view_show(XePUIMessageView *mview, 
				   XePMessage *message)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");

}

void
a_TtyUI_messageview_mime_view_block(XePUIMessageView *mview, 
				    XePMessage *message)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");

}

void
a_TtyUI_messageview_mime_view_unblock(XePUIMessageView *mview,
				      XePMessage *message)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");

}

void
a_TtyUI_messageview_mime_show_selected(XePUIMessageView *mview, XePMessage *message)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");
}

XePUIComposeWindow *
a_TtyUI_composewindow_create(XetPan *xetpan, char *terminal)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");

  return NULL;
}

void
a_TtyUI_composewindow_set_message(XePUIComposeWindow *uicompose, 
				  XePMessage *message)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");
}

void
a_TtyUI_composewindow_set_body_text(XePUIComposeWindow *uicompose,
				    const char *text,
				    size_t text_size,
				    const char *text_charset)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");
}

void
a_TtyUI_composewindow_add_header(XePUIComposeWindow *uicompose,
				 XePMessageHeaderType header_type,
				 const char *header_name,
				 const char *header_value)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");
}

void
a_TtyUI_composewindow_close(XePUIComposeWindow *uicompose)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");
}

XePUIMessageView*
a_TtyUI_messageview_create_standalone(XetPan *xetpan, 
				      char *terminal)
{
 A_LOGGING_SYSTEM_ERROR("not implemented yet");

  return NULL;
}

void
a_TtyUI_messageview_close(XePUIMessageView *mview)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");
}

void
a_TtyUI_about(XePUIMain *uimain)
{
  A_LOGGING_SYSTEM_ERROR("not implemented yet");
}

void
a_TtyUI_notify_error(XePUIMain *uimain)
{
  flash();
}

void
a_TtyUI_notify_error_reset(XePUIMain *uimain)
{

}

static void
TtyUI_python_console(XePUIMain *uimain)
{
  XePTTYPython *xeppy = (XePTTYPython *) uimain->python;
  if (xeppy == NULL)
    {
      uimain->python = TtyPython_console_new(uimain->xetpan,
					       NULL);
    }
  else
    {
      TtyPython_console_show(xeppy);
    }
}


static XePTTYPython *
TtyPython_console_new(XetPan *xetpan, char *terminal)
{
  /* Create a new window and launch Python
   * The main XetPan loop should be used so 
   */
  A_LOGGING_TRACE;
  printf("Python console..\n");
  return NULL;
}

static void
TtyPython_console_show(XePTTYPython *xeppy)
{
  A_LOGGING_TRACE;
}

/**********************************************************************/
/*                                                                    */
/* I/O traffic watchers                                               */
/*                                                                    */
/**********************************************************************/

int
a_TtyUI_create_io_watch(XePUI *ui,
			XePUIIOWatcher *watcher)
{
  XePTTYUI *ttyui = ui->tty;
  struct pollfd *pfd;

  A_LOGGING_TRACE;

  if (ttyui->nfds == ttyui->fds_size)
    {
      struct pollfd *newfds;
      int new_size;

      A_LOGGING_DEBUG("resizing pollfd array");
      new_size = ttyui->fds_size * 2;
      newfds = (struct pollfd *) realloc(ttyui->fds, 
					 new_size * sizeof(struct pollfd));

      if (newfds == NULL)
	{
	  A_LOGGING_SYSTEM_ERROR("realloc: %s", strerror(errno));
	  return -1;
	}
      ttyui->fds = newfds;
      ttyui->fds_size = new_size;
    }

  pfd = &(ttyui->fds[ttyui->nfds]);
  pfd->fd = watcher->fd;
  pfd->events = TtyUI_condition_set(watcher->condition);
  
  watcher->channel = (void *) (ttyui->nfds);

  ttyui->nfds ++;

  TtyUI_io_info(ui);
  return 0;
}

void
a_TtyUI_delete_io_watch(XePUI *ui, XePUIIOWatcher *watcher)
{
  XePTTYUI *ttyui = ui->tty;
  int old_pfd_idx;
  int last_pfd_idx;

  /* old_pfd_idx contains the index where the removed watcher _was_ In
   * that index is now the watcher that _was_ the last one in
   * the watcher list ui->watched_fd. This is because carray_delete
   * that is used in a_UI_remove_io_watch replaces a removed item by
   * the last item in the array.
   */

  A_LOGGING_TRACE;

  if (ttyui->nfds > 0)
    {
      A_LOGGING_DEBUG("id %d fd %d condition 0x%x channel %d",
		      watcher->watch_id,
		      watcher->fd,
		      watcher->condition,
		      (int) watcher->channel);
      
      old_pfd_idx = (int) watcher->channel;
      watcher->channel = (void *) -1; /* Invalidate watcher channel */

      last_pfd_idx = ttyui->nfds - 1;


      if (old_pfd_idx < last_pfd_idx)
        {
          struct pollfd *pfd;
          XePUIIOWatcher *former_last_watcher;
          
          A_LOGGING_DEBUG("replacing pfd %d by pfd %d",
			  old_pfd_idx,
			  last_pfd_idx);

          /* WARNING: when switching UI, the watchers are not deleted!!! */
          /* Move the last pfd to the place of the removed one.
           * This is done because carray_delete used in a_UI_remove_io_watch
           * works like this. Ugly, but necessary.
           * TODO: find a cleaner approach 
           */
          pfd = &(ttyui->fds[old_pfd_idx]);
          pfd->fd     = (ttyui->fds[last_pfd_idx]).fd;
          pfd->events = (ttyui->fds[last_pfd_idx]).events;
      
          /* last_last_watcher is the watcher that was the last one in
           * the ui->watched_fd list before the watcher was removed by
           * carray_delete in a_UI_remove_io_watch. Ugly but necessary.
           * TODO: find a cleaner approach
           */
          former_last_watcher = (XePUIIOWatcher *) carray_get(ui->watched_fd,
                                                              old_pfd_idx);
          former_last_watcher->channel = (void *) old_pfd_idx;
          A_LOGGING_DEBUG("watcher id %d has now channel %d",
			  former_last_watcher->watch_id, 
                         (int ) former_last_watcher->channel);
        }
  
      ttyui->nfds = last_pfd_idx;
      TtyUI_io_info(ui);
    }
  else
    {
      A_LOGGING_SYSTEM_ERROR("registered fd to poll");
    }

}
static void
TtyUI_io_info(XePUI *ui)
{
  XePTTYUI *ttyui = (XePTTYUI *) ui->tty;
  XePUIIOWatcher *watcher;
  struct pollfd *pfd;
  int i;
  
  A_LOGGING_TRACE;
  
  fprintf(stderr, "\t tty->nfds = %d watched_fd count = %d\n",
	  ttyui->nfds, carray_count(ui->watched_fd));
  fprintf(stderr, "chan\tf-fd\tf-ev\tw-id\tw-fd\tw-cond\tw-chan\n");
  for (i = 0; i < ttyui->nfds; i++)
    {
      pfd = &(ttyui->fds[i]);
      watcher = carray_get(ui->watched_fd, i);
      fprintf(stderr, "%d\t%d\t%2x\t%d\t%d\t%2x\t%d\n", i,
	      pfd->fd, pfd->events, watcher->watch_id, watcher->fd,
	      watcher->condition, (int )watcher->channel);
    }
}
#endif /* USE_TTYUI */
